from unittest.mock import MagicMock

from mambu_tests.constants import config_json
from mambu_tests.tap_generators import setup_generator_base_test
from tap_mambu.helpers.generator_processor_pairs import get_available_streams
from tap_mambu.tap_generators.no_pagination_generator import NoPaginationGenerator


def test_no_pagination_generators():
    client_mock = MagicMock()
    client_mock.page_size = int(config_json.get("page_size", 1))

    for stream in get_available_streams():
        generators = setup_generator_base_test(stream, client_mock=client_mock)

        for generator in generators:
            if not isinstance(generator, NoPaginationGenerator):
                continue
            return_values = list()

            client_mock.request = MagicMock()
            client_mock.request.return_value = [{} for _ in range(2)]

            for val in generator:
                return_values.append(val)
                client_mock.request.assert_called_once_with(
                    method=generator.endpoint_api_method,
                    path=generator.endpoint_path,
                    version=generator.endpoint_api_version,
                    apikey_type=generator.endpoint_api_key_type,
                    params="",
                    endpoint=stream,
                    json={"sortingCriteria": {},
                          "filterCriteria": []}
                )
            assert len(client_mock.request.return_value) == len(return_values)



def convert_snake_to_pascal(snake_case):
    return ''.join(word.capitalize() if len(word) > 2 else word.upper() for word in snake_case.split('_'))


def convert_pascal_to_snake(pascal_case):
    return ''.join(f'_{c.lower()}' if c.isupper() else c for c in pascal_case.replace('ID', '_id')).lstrip('_')



from . import setup_generator_base_test


def test_loan_products_generator_endpoint_config_init():
    generators = setup_generator_base_test("loan_products")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == "loanproducts"
    assert generator.endpoint_api_method == "GET"
    assert generator.endpoint_params.get("sortBy") == "id:ASC"
    assert generator.endpoint_bookmark_field == "lastModifiedDate"



from .multithreaded_bookmark_generator import MultithreadedBookmarkGenerator
from ..helpers import get_bookmark
from ..helpers.datetime_utils import datetime_to_utc_str, str_to_localized_datetime


class LoanTransactionsGenerator(MultithreadedBookmarkGenerator):
    def _init_endpoint_config(self):
        super(LoanTransactionsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "loans/transactions:search"
        self.endpoint_bookmark_field = "creationDate"
        self.endpoint_sorting_criteria["field"] = "id"
        self.endpoint_filter_criteria = [
            {
                "field": "creationDate",
                "operator": "AFTER",
                "value": datetime_to_utc_str(str_to_localized_datetime(
                    get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)))
            }
        ]

    def prepare_batch_params(self):
        super(LoanTransactionsGenerator, self).prepare_batch_params()
        self.endpoint_filter_criteria[0]["value"] = datetime_to_utc_str(self.endpoint_intermediary_bookmark_value)



"""
Test that the tap respects the start date
"""
from singer.utils import strptime_to_utc
from tap_tester import connections, menagerie, runner
from base import MambuBaseTest

class StartDateTest(MambuBaseTest):
    """
    Test that the tap respects the start date
    """

    first_sync_start_date = None
    second_sync_start_date = None
    first_sync_records = None
    second_sync_records = None

    @staticmethod
    def name():
        return "tap_tester_mambu_start_date_test"

    def setUp(self):
        self.first_sync_start_date = self.get_properties()['start_date']

        self.second_sync_start_date = self.get_properties(
            original_properties=False
        )['start_date']

    def untestable_streams(self):
        return set([
            "communications", # Need to set up Twilio or email server to send stuff
            "installments",
        ])

    def get_replication_key_values(self, stream, records):
        all_records = []
        for record in records:
            # Build the primary key for this record, maintaining the same order for the fields
            record_rep_key = [record[field]
                              for field in sorted(self.expected_replication_keys()[stream]) if field in record]
            # Cast to a tuple to make it hashable
            all_records.append(record_rep_key[0])
        return all_records

    def test_run(self):
        """
        Verify that running a sync, then moving the start date into the
        future and running a sync results in less records than the first
        sync
        """

        conn_id = connections.ensure_connection(self)
        self.run_and_verify_check_mode(conn_id)

        self.select_and_verify_fields(conn_id)

        first_sync_record_count_by_stream = self.run_and_verify_sync(conn_id)

        first_sync_state = menagerie.get_state(conn_id)
        first_sync_all_records_by_stream = runner.get_records_from_target_output()


        conn_id = connections.ensure_connection(self, original_properties = False)
        self.run_and_verify_check_mode(conn_id)

        self.select_and_verify_fields(conn_id)

        second_sync_record_count_by_stream = self.run_and_verify_sync(conn_id)

        second_sync_state = menagerie.get_state(conn_id)
        second_sync_all_records_by_stream = runner.get_records_from_target_output()

        all_metadata = self.expected_metadata()
        for stream_name in self.expected_sync_streams():
            stream_metadata = all_metadata[stream_name]
            with self.subTest(stream=stream_name):

                replication_method = stream_metadata.get(self.REPLICATION_METHOD)

                first_sync_count = first_sync_record_count_by_stream.get(stream_name,0)
                second_sync_count = second_sync_record_count_by_stream.get(stream_name,0)

                first_sync_records = []
                for message in first_sync_all_records_by_stream[stream_name]['messages']:
                    if message['action'] == 'upsert':
                        first_sync_records.append(message['data'])

                second_sync_records = []
                for message in second_sync_all_records_by_stream[stream_name]['messages']:
                    if message['action'] == 'upsert':
                        second_sync_records.append(message['data'])

                if  replication_method == self.FULL_TABLE:
                    """
                    Testing Criteria:
                    1. Verify that Sync B includes the same number of records as Sync A
                    2. Verify the saved state following each sync does not contain a bookmark value
                       for the stream.
                    3. Verify that all records in Sync B are included in Sync A.
                    """
                    # Criteria 1
                    self.assertGreaterEqual(first_sync_count, second_sync_count)

                    # Criteria 2
                    self.assertNotIn(stream_name, first_sync_state['bookmarks'])
                    self.assertNotIn(stream_name, second_sync_state['bookmarks'])

                    # Criteria 3
                    # - Note: This may cover up the scenario where we sync
                    #   duplicate records, but we check for that in a
                    #   different test. So we ignore that here
                    first_sync_unique_records = self.get_unique_records(stream_name,
                                                                        first_sync_records)
                    second_sync_unique_records = self.get_unique_records(stream_name,
                                                                         second_sync_records)
                    self.assertGreaterEqual(len(first_sync_unique_records),
                                            len(second_sync_unique_records))
                else:
                    """
                    Testing Criteria:
                    1. Verify the saved state following each sync contains an appropriate bookmark
                       value that corresponds to the stream’s replication key.
                    2. Verify the record count for each stream is greater in Sync A than Sync B.
                    3. Verify all records in Sync A and Sync B have replication key values which are
                       greater than or equal to the corresponding start date for that sync.
                    """
                    # Criteria 1
                    self.assertGreaterEqual(first_sync_count, second_sync_count)

                    # Criteria 2
                    self.assertIn(stream_name, first_sync_state['bookmarks'])
                    self.assertIn(stream_name, second_sync_state['bookmarks'])

                    # Criteria 3
                    rep_values = self.get_replication_key_values(stream_name, first_sync_records)

                    for value in rep_values:
                        self.assertGreaterEqual(
                            strptime_to_utc(value),
                            strptime_to_utc(self.first_sync_start_date)
                        )

                    rep_values = self.get_replication_key_values(stream_name, second_sync_records)

                    for value in rep_values:
                        self.assertGreaterEqual(
                            strptime_to_utc(value),
                            strptime_to_utc(self.second_sync_start_date)
                        )



from .generator import TapGenerator


class ClientsGenerator(TapGenerator):
    def _init_endpoint_config(self):
        super(ClientsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "clients"
        self.endpoint_api_method = "GET"



"""
Test that with no fields selected for a stream automatic fields are still replicated
"""
from tap_tester import connections
from base import MambuBaseTest

class SyncCanaryTest(MambuBaseTest):
    """
    Smoke test
    """

    @staticmethod
    def name():
        return "tap_tester_mambu_sync_canary_test"

    def test_run(self):
        """
        Run tap in check mode, then select all streams and all fields within streams. Run a sync and
        verify exit codes do not throw errors. This is meant to be a smoke test for the tap. If this
        is failing do not expect any other tests to pass.
        """
        conn_id = connections.ensure_connection(self)
        self.run_and_verify_check_mode(conn_id)

        self.select_and_verify_fields(conn_id)

        record_count_by_stream = self.run_and_verify_sync(conn_id)


        # Assert all expected streams synced at least one record
        for stream in self.expected_streams():
            with self.subTest(stream=stream):
                self.assertGreater(record_count_by_stream.get(stream, 0),
                                   0,
                                   msg="{} did not sync any records".format(stream))



from datetime import datetime

from singer import utils

from . import setup_generator_base_test


def test_gl_journal_entries_generator_endpoint_config_init():
    generators = setup_generator_base_test("gl_journal_entries")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == 'gljournalentries:search'
    assert generator.endpoint_api_method == "POST"
    assert generator.endpoint_params == {
        "detailsLevel": "FULL",
        "paginationDetails": "OFF"
    }
    assert generator.endpoint_sorting_criteria == {
            "field": "entryId",
            "order": "ASC"
        }
    assert generator.endpoint_filter_criteria[0] == {
        "field": "creationDate",
        "operator": "BETWEEN",
        "value": "2021-06-01T00:00:00.000000Z",
        "secondValue": datetime.strptime(generator.endpoint_filter_criteria[0]["secondValue"],
                                         "%Y-%m-%dT%H:%M:%S.%fZ").strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    }
    assert generator.endpoint_bookmark_field == "creationDate"



from . import setup_generator_base_test


def test_loan_accounts_generator():
    generators = setup_generator_base_test("loan_accounts")

    assert 2 == len(generators)

    lm_generator = generators[0]

    assert lm_generator.endpoint_path == "loans:search"
    assert lm_generator.endpoint_bookmark_field == "lastModifiedDate"
    assert lm_generator.endpoint_sorting_criteria == {
            "field": "id",
            "order": "ASC"
        }
    assert lm_generator.endpoint_filter_criteria == [
            {
                "field": "lastModifiedDate",
                "operator": "AFTER",
                "value": '2021-05-31T17:00:00.000000Z'
            }
        ]

    ad_generator = generators[1]

    assert ad_generator.endpoint_path == "loans:search"
    assert ad_generator.endpoint_bookmark_field == "lastAccountAppraisalDate"
    assert ad_generator.endpoint_sorting_criteria == {
            "field": "id",
            "order": "ASC"
        }
    assert ad_generator.endpoint_filter_criteria == [
        {
            "field": "lastAccountAppraisalDate",
            "operator": "AFTER",
            "value": '2021-05-31T17:00:00.000000Z'
        }
    ]



from .generator import TapGenerator


class CentresGenerator(TapGenerator):
    def _init_endpoint_config(self):
        super(CentresGenerator, self)._init_endpoint_config()
        self.endpoint_path = "centres"
        self.endpoint_api_method = "GET"
        self.endpoint_params = {
            "sortBy": "lastModifiedDate:ASC",
            "detailsLevel": "FULL",
            "paginationDetails": "OFF"
        }
        self.endpoint_bookmark_field = "lastModifiedDate"



from singer.transform import Transformer as SingerTransformer

from tap_mambu.helpers.datetime_utils import datetime_to_utc_str, str_to_localized_datetime


class Transformer(SingerTransformer):
    def _transform_datetime(self, value):
        return datetime_to_utc_str(str_to_localized_datetime(value))


def transform_datetime(this_dttm):
    with Transformer() as transformer:
        # noinspection PyProtectedMember
        # pylint: disable=W0212
        new_dttm = transformer._transform_datetime(this_dttm)
    return new_dttm



"""
Test that the tap can replicate multiple pages of data
"""
from datetime import timedelta
import backoff
from singer.utils import strftime, strptime_to_utc
from tap_tester import connections, menagerie, runner
from base import MambuBaseTest

@backoff.on_predicate(backoff.expo, lambda x: x <= 0, max_tries=10)
def poll_state_version(conn_id):
    """Make the request for state version until it returns a version greater than 0"""
    return menagerie.get_state_version(conn_id)

class BookmarksTest(MambuBaseTest):
    """
    Test that the tap can replicate multiple pages of data
    """

    @staticmethod
    def name():
        return "tap_tester_mambu_bookmarks_test"

    def untestable_streams(self):
        return set([
            "communications", # Need to set up Twilio or email server to send stuff
            "installments",
            "gl_accounts",
        ])

    def subtract_day(self, bookmark):
        bookmark_dt = strptime_to_utc(bookmark)
        adjusted_bookmark = bookmark_dt - timedelta(days=1)
        return strftime(adjusted_bookmark)

    def test_run(self):
        """
        Verify that we can get multiple pages of data for each stream
        """
        conn_id = connections.ensure_connection(self)
        self.run_and_verify_check_mode(conn_id)

        self.select_and_verify_fields(conn_id)

        first_sync_record_count = self.run_and_verify_sync(conn_id)

        first_sync_bookmarks = menagerie.get_state(conn_id)
        first_sync_records = runner.get_records_from_target_output()


        new_bookmarks = {}
        for stream_name, current_bookmark in first_sync_bookmarks['bookmarks'].items():
            if stream_name == 'gl_accounts':
                new_gl_bookmarks = {
                    sub_stream : self.subtract_day(sub_bookmark)
                    for sub_stream, sub_bookmark in current_bookmark.items()
                }
                new_bookmarks[stream_name] = new_gl_bookmarks
            else:
                new_bookmarks[stream_name] = self.subtract_day(current_bookmark)

        new_state = {"bookmarks" : new_bookmarks}

        # Ensure the test is not the first to post a state
        poll_state_version(conn_id)

        menagerie.set_state(conn_id, new_state)

        # Run a sync job using orchestrator
        second_sync_record_count = self.run_and_verify_sync(conn_id)
        second_sync_bookmarks = menagerie.get_state(conn_id)
        second_sync_records = runner.get_records_from_target_output()

        for stream in self.expected_sync_streams():
            with self.subTest(stream=stream):
                replication_method = self.expected_replication_method().get(stream)

                first_sync_count = first_sync_record_count.get(stream, 0)
                second_sync_count = second_sync_record_count.get(stream, 0)

                first_sync_messages = first_sync_records.get(stream, {}).get('messages', [])
                second_sync_messages = second_sync_records.get(stream, {}).get('messages', [])

                if replication_method == self.INCREMENTAL:
                    replication_keys = self.expected_replication_keys().get(stream)

                    first_sync_bookmark_value = first_sync_bookmarks['bookmarks'][stream]
                    second_sync_bookmark_value = second_sync_bookmarks['bookmarks'][stream]
                    simulated_bookmark_value = new_state['bookmarks'][stream]

                    # Verify the both syncs end on the same bookmark
                    self.assertEqual(first_sync_bookmark_value,
                                     second_sync_bookmark_value)

                    # Verify that first sync records fall between the start date and the final
                    # bookmark value
                    for message in first_sync_messages:
                        lower_bound = strptime_to_utc(self.get_properties()['start_date'])
                        upper_bound = strptime_to_utc(first_sync_bookmark_value)
                        record = message.get('data')
                        actual_values = [strptime_to_utc(record.get(replication_key))
                                         for replication_key in replication_keys if replication_key in record]

                        self.assertNotEqual(actual_values, [], msg="No replication key found in record")

                        self.assertTrue(
                            any([lower_bound <= actual_value <= upper_bound for actual_value in actual_values]),
                            msg="First sync records fall outside of expected sync window"
                        )

                    # Verify the second sync records fall between simulated bookmark value and the
                    # final bookmark value
                    for message in second_sync_messages:
                        lower_bound = strptime_to_utc(simulated_bookmark_value)
                        upper_bound = strptime_to_utc(second_sync_bookmark_value)
                        record = message.get('data')
                        actual_values = [strptime_to_utc(record.get(replication_key))
                                         for replication_key in replication_keys if replication_key in record]

                        self.assertNotEqual(actual_values, [], msg="No replication key found in record")

                        self.assertTrue(
                            any([lower_bound <= actual_value <= upper_bound for actual_value in actual_values]),
                            msg="Second sync records fall outside of expected sync window"
                        )

                    # Verify the number of records in the 2nd sync is less then the first
                    self.assertLess(second_sync_count, first_sync_count)

                    # Verify at least 1 record was replicated in the second sync
                    self.assertGreater(
                        second_sync_count,
                        0,
                        msg="We are not fully testing bookmarking for {}".format(stream)
                    )

                elif replication_method == self.FULL_TABLE:
                    # Verify no bookmark exists
                    self.assertNotIn(stream, first_sync_bookmarks['bookmarks'])
                    self.assertNotIn(stream, second_sync_bookmarks['bookmarks'])

                else:
                    raise NotImplementedError(
                        "invalid replication method: {}".format(replication_method)
                    )



from .multithreaded_bookmark_generator import MultithreadedBookmarkGenerator
from ..helpers import get_bookmark
from ..helpers.datetime_utils import datetime_to_local_str, str_to_localized_datetime


class CommunicationsGenerator(MultithreadedBookmarkGenerator):
    def _init_endpoint_config(self):
        super(CommunicationsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "communications/messages:search"
        self.endpoint_params = {
            "detailsLevel": "FULL",
            "paginationDetails": "OFF"
        }
        self.endpoint_bookmark_field = "creationDate"
        self.endpoint_filter_criteria = [
            {
                "field": "state",
                "operator": "EQUALS",
                "value": "SENT"
            },
            {
                "field": "creationDate",
                "operator": "AFTER",
                "value": datetime_to_local_str(str_to_localized_datetime(
                    get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)))
            }
        ]

    def _init_endpoint_body(self):
        self.endpoint_body = self.endpoint_filter_criteria

    def prepare_batch_params(self):
        super(CommunicationsGenerator, self).prepare_batch_params()
        self.endpoint_filter_criteria[1]["value"] = datetime_to_local_str(self.endpoint_intermediary_bookmark_value)



from mock import MagicMock

from tap_mambu.tap_generators.child_generator import ChildGenerator
from ..constants import config_json
from tap_mambu.helpers.generator_processor_pairs import get_generator_processor_for_stream
from ..helpers import ClientWithDataMock, ClientMock


def setup_generator_base_test(stream_name, client_mock=None, with_data=False, custom_data=None,
                              offset_field="offset", limit_field="limit", bookmark_field="creationDate"):
    generator_classes, _ = get_generator_processor_for_stream(stream_name)
    generators = list()
    for generator_class in generator_classes:
        client = client_mock
        if client_mock is None:
            client = ClientMock(int(config_json.get("page_size", 200)))
            if with_data:
                client = ClientWithDataMock(int(config_json.get("page_size", 200)), custom_data=custom_data,
                                            offset_field=offset_field, limit_field=limit_field,
                                            bookmark_field=bookmark_field)

        generator = generator_class(stream_name=stream_name,
                                    client=client,
                                    config=config_json,
                                    state={"currently_syncing": stream_name},
                                    sub_type="self",
                                    **({"parent_id": "TEST"} if issubclass(generator_class, ChildGenerator) else {}))
        generators.append(generator)
    return generators



from . import setup_processor_base_test


def test_loan_accounts_processor():
    processor = setup_processor_base_test("loan_accounts")

    assert processor.endpoint_deduplication_key == "id"
    assert processor.endpoint_child_streams == ["loan_repayments"]



"""
Test that when no fields are selected for a stream, automatic fields are still replicated
"""

from tap_tester import connections, runner
from base import MambuBaseTest

class PaginationTest(MambuBaseTest):
    """
    Test that the tap can paginate API endpoints
    """

    @staticmethod
    def name():
        return "tap_tester_mambu_pagination_test"

    def untestable_streams(self):
        return set([
            "communications", # Need to set up Twilio or email server to send stuff
        ])

    def test_run(self):
        """
        Verify that we can get multiple pages of unique records for each
        stream
        """

        conn_id = connections.ensure_connection(self)
        self.run_and_verify_check_mode(conn_id)

        self.select_and_verify_fields(conn_id)

        record_count_by_stream = self.run_and_verify_sync(conn_id)

        all_records_by_stream = runner.get_records_from_target_output()
        page_size = int(self.get_properties()['page_size'])

        for stream in self.expected_sync_streams():
            with self.subTest(stream=stream):
                # Assert all expected streams synced at least a full pages of records
                self.assertGreater(
                    record_count_by_stream.get(stream, 0),
                    page_size,
                    msg="{} did not sync more than a page of records".format(stream)
                )

                records = [ x['data'] for x in all_records_by_stream[stream]['messages']]

                unique_records = self.get_unique_records(stream, records)

                self.assertGreater(len(unique_records),
                                   page_size)



"""
Test that when no fields are selected for a stream, automatic fields are still replicated
"""
from tap_tester import runner, connections
from base import MambuBaseTest

class AutomaticFieldsTest(MambuBaseTest):
    """
    Test that when no fields are selected for a stream, automatic fields
    are still replicated
    """

    @staticmethod
    def name():
        return "tap_tester_mambu_automatic_fields_test"

    def untestable_streams(self):
        return set([
            "communications", # Need to set up Twilio or email server to send stuff
        ])

    def verify_is_selected(self, is_selected):
        """
        Override the default `verify_is_selected` because we only want automatic fields selected
        """
        self.assertFalse(is_selected)

    def test_run(self):
        """
        Verify that we can get multiple pages of automatic fields for each
        stream
        """

        conn_id = connections.ensure_connection(self)
        self.run_and_verify_check_mode(conn_id)

        self.select_and_verify_fields(conn_id, select_all_fields = False)

        record_count_by_stream = self.run_and_verify_sync(conn_id)

        actual_fields_by_stream = runner.examine_target_output_for_fields()

        # Assert all expected streams synced at least a full pages of records
        for stream in self.expected_sync_streams():
            with self.subTest(stream=stream):
                self.assertGreater(record_count_by_stream.get(stream, 0),
                                   int(self.get_properties()['page_size']),
                                   msg="{} did not sync more than a page of records".format(stream))

        for stream_name, actual_fields in actual_fields_by_stream.items():
            with self.subTest(stream=stream_name):
                self.assertSetEqual(self.expected_automatic_fields()[stream_name],
                                    actual_fields)



"""
Test tap discovery
"""
import re
from tap_tester import connections, menagerie
from base import MambuBaseTest

class DiscoveryTest(MambuBaseTest):
    """ Test the tap discovery """

    @staticmethod
    def name():
        return "tap_tester_mambu_discovery_test"

    def test_run(self):
        """
        • Verify that discover creates the appropriate catalog, schema, metadata, etc.
        • Verify number of actual streams discovered match expected
        • Verify the stream names discovered were what we expect
        • Verify stream names follow naming convention
          • streams should only have lowercase alphas and underscores
        • verify there is only 1 top level breadcrumb
        • verify replication key(s)
        • verify primary key(s)
        • verify that if there is a replication key we are doing INCREMENTAL otherwise FULL_TABLE
        • verify the actual replication matches our expected replication method
        • verify that primary and replication keys are given the inclusion of automatic metadata
        • verify that all other fields have inclusion of available metadata
        """
        streams_to_test = self.expected_streams()

        conn_id = connections.ensure_connection(self)
        self.run_and_verify_check_mode(conn_id)

        # Verify number of actual streams discovered match expected
        catalogs = menagerie.get_catalogs(conn_id)
        found_catalogs = [catalog for catalog in catalogs
                          if catalog.get('tap_stream_id') in streams_to_test]

        self.assertGreater(len(found_catalogs),
                           0,
                           msg="unable to locate schemas for connection {}".format(conn_id))
        self.assertEqual(len(streams_to_test),
                         len(found_catalogs),
                         msg="Expected {} streams, actual was {} for connection {}".format(
                             len(streams_to_test),
                             len(found_catalogs),
                             conn_id)
                         )

        # Verify the stream names discovered were what we expect
        found_catalog_names = {c['tap_stream_id'] for c in found_catalogs}
        self.assertEqual(set(streams_to_test),
                         set(found_catalog_names),
                         msg="Expected streams don't match actual streams")

        # Verify stream names follow naming convention
        # streams should only have lowercase alphas and underscores
        self.assertTrue(all([re.fullmatch(r"[a-z_]+", name) for name in found_catalog_names]),
                        msg="One or more streams don't follow standard naming")

        for stream in streams_to_test:
            with self.subTest(stream=stream):
                catalog = next(iter([catalog for catalog in found_catalogs
                                     if catalog["stream_name"] == stream]))
                assert catalog  # based on previous tests this should always be found

                schema_and_metadata = menagerie.get_annotated_schema(conn_id, catalog['stream_id'])
                metadata = schema_and_metadata["metadata"]

                # verify the stream level properties are as expected

                # verify there is only 1 top level breadcrumb
                stream_properties = [item for item in metadata if item.get("breadcrumb") == []]
                stream_metadata = stream_properties[0].get("metadata", {})
                self.assertTrue(len(stream_properties) == 1,
                                msg=("There is NOT only one top level breadcrumb for {}"
                                "\nstream_properties | {}").format(stream, stream_properties))

                # verify replication key(s)
                expected = self.expected_replication_keys()[stream]
                actual = set(stream_metadata.get(self.REPLICATION_KEYS, []))

                self.assertEqual(
                    expected,
                    actual,
                    msg="expected replication key {} but actual is {}".format(expected, actual))

                # verify primary key(s)
                expected = self.expected_primary_keys()[stream]
                actual = set(stream_metadata.get(self.PRIMARY_KEYS, []))

                self.assertEqual(
                    expected,
                    actual,
                    msg="expected primary key {} but actual is {}".format(expected, actual))

                # verify that if there is a replication key we are doing INCREMENTAL otherwise FULL


                actual_replication_method = stream_metadata.get(self.REPLICATION_METHOD)

                if stream_metadata.get(self.REPLICATION_KEYS):
                    self.assertTrue(
                        self.INCREMENTAL == actual_replication_method,
                        msg="Expected INCREMENTAL replication since there is a replication key")
                else:
                    self.assertTrue(
                        self.FULL_TABLE == actual_replication_method,
                        msg="Expected FULL_TABLE replication since there is no replication key")

                # verify the actual replication matches our expected replication method
                expected_replication_method = self.expected_replication_method().get(stream, None)
                self.assertEqual(
                    expected_replication_method,
                    actual_replication_method,
                    msg="expected replication method {} but actual is {}".format(
                        expected_replication_method,
                        actual_replication_method))

                expected_primary_keys = self.expected_primary_keys()[stream]
                expected_replication_keys = self.expected_replication_keys()[stream]
                expected_automatic_fields = expected_primary_keys | expected_replication_keys

                # verify that primary and replication keys are given `"inclusion": "automatic"`
                # metadata.

                for item in metadata:
                    # Skip the stream level metadata
                    if item.get("breadcrumb") == []:
                        continue

                    actual_field_inclusion = item.get("metadata", {}).get("inclusion")
                    field_name = item.get("breadcrumb", ["properties", None])[1]

                    if field_name:
                        if actual_field_inclusion == "automatic":
                            self.assertIn(field_name, expected_automatic_fields)
                        elif actual_field_inclusion == "available":
                            self.assertNotIn(field_name, expected_automatic_fields)
                        else:
                            raise RuntimeError("Stream {} got unexpected `inclusion` value {}"
                                               .format(stream, field_name))
                    else:
                        raise RuntimeError("Got null field name on stream {}".format(stream))



from .generator import TapGenerator


class CreditArrangementsGenerator(TapGenerator):
    def _init_endpoint_config(self):
        super(CreditArrangementsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "creditarrangements"
        self.endpoint_api_method = "GET"
        self.endpoint_params["sortBy"] = "creationDate:ASC"
        self.endpoint_bookmark_field = "lastModifiedDate"



from .deduplication_processor import DeduplicationProcessor
from .multithreaded_parent_processor import MultithreadedParentProcessor


class LoanAccountsProcessor(MultithreadedParentProcessor, DeduplicationProcessor):
    def _init_endpoint_config(self):
        super(LoanAccountsProcessor, self)._init_endpoint_config()
        self.endpoint_deduplication_key = "id"
        self.endpoint_child_streams = ["loan_repayments"]



import os
import shutil
import json
from schema_fetcher.constants import OUTPUT_DIR_PATH, RESOURCES_LINK, CUSTOM_FIELDS_FIELD
from schema_fetcher.converters import convert_pascal_to_snake, convert_snake_to_pascal
from schema_fetcher.helpers import get_mambu_client, get_data_type_and_format
from schema_fetcher.custom_exceptions import ResourceFileNotFound, StreamJsonObjectNotFound


class Streams:
    def __init__(self):
        self.__swaggered_streams = {
            'gl_accounts': 'general_ledger_accounts',
            'gl_journal_entries': 'journal_entries',
            'interest_accrual_breakdown': 'accounting_interest_accrual',
        }

        self.__tap_streams_singular_form = {
            'branches': 'branch',
            'communications': 'communication',
            'centres': 'centre',
            'clients': 'client',
            'credit_arrangements': 'credit_arrangement',
            'custom_field_sets': 'custom_field_set',
            'deposit_accounts': 'deposit_account',
            'deposit_products': 'deposit_product',
            'deposit_transactions': 'deposit_transaction',
            'groups': 'group',
            'loan_accounts': 'loan_account',
            'loan_products': 'loan_product',
            'loan_transactions': 'loan_transaction',
            'tasks': 'task',
            'users': 'user',
            'gl_accounts': 'gl_account',
            'gl_journal_entries': 'gl_journal_entry',
            'index_rate_sources': 'index_rate_source',
            'installments': 'installment',
            'interest_accrual_breakdown': 'interest_accrual_breakdown',
        }

        self.__stream_json_obj = {
            'communications': 'communication_message',
            'custom_field_sets': 'custom_field_meta',
        }

    def get_stream_json_obj_form(self, stream_name):
        stream_json_obj = dict(self.__tap_streams_singular_form, **self.__stream_json_obj)

        if stream_name in self.__swaggered_streams.values():
            for tap_stream, swaggered_stream in self.__swaggered_streams.items():
                if swaggered_stream == stream_name:
                    return stream_json_obj[tap_stream]
        return stream_json_obj[stream_name]

    def get_stream_names_swaggered(self):
        streams_singular_form = list(self.__tap_streams_singular_form.keys())
        for idx, val in enumerate(streams_singular_form):
            streams_singular_form[idx] = self.__swaggered_streams.get(val, val)
        return streams_singular_form

    def convert_swaggered_to_tap_stream(self, swaggered_stream):
        if swaggered_stream in self.__swaggered_streams.values():
            for tap_stream, swaggered in self.__swaggered_streams.items():
                if swaggered == swaggered_stream:
                    return tap_stream
        return swaggered_stream

    def get_tap_streams_count(self):
        return len(self.__tap_streams_singular_form)

    def get_tap_streams_name(self):
        return self.__tap_streams_singular_form.keys()


streams = Streams()
client = get_mambu_client()


def get_stream_resource_file_paths():
    available_resources = client.request(method='GET',
                                         path=RESOURCES_LINK,
                                         version='remove_accept_header')
    file_paths = []
    found_streams_swagger_format = []
    swagger_format_streams = []

    tap_streams_swaggered = streams.get_stream_names_swaggered()
    for file_data in available_resources['items']:
        stream_name = file_data['hashValue'].lower()
        swagger_format_streams.append(stream_name)
        if stream_name in tap_streams_swaggered:
            found_streams_swagger_format.append(stream_name)
            file_paths.append(file_data['jsonPath'])

    if len(found_streams_swagger_format) != len(streams.get_stream_names_swaggered()):
        raise ResourceFileNotFound(f'Resource file not found for the following streams: '
                                   f'{", ".join([stream for stream in streams.get_stream_names_swaggered() if stream not in found_streams_swagger_format])}\n'
                                   f'Available swagger streams: {", ".join(swagger_format_streams)}')

    return zip(found_streams_swagger_format, file_paths)


def generate_json_objs(field_properties, all_schema_obj, type_object_nullable=True):
    full_schema = {**get_data_type_and_format('object', obj_type_nullable=type_object_nullable),
                   'additionalProperties': False,
                   'properties': {}}

    for field_name, field_prop in field_properties['properties'].items():
        field_name = convert_pascal_to_snake(field_name)
        if 'items' in field_prop:
            full_schema['properties'][field_name] = {'anyOf': [{**get_data_type_and_format(field_prop['type'],
                                                                                           field_prop.get('format',
                                                                                                          None)),
                                                                'items': {}},
                                                               {'type': 'null'}]}
            if 'originalRef' in field_prop['items']:
                full_schema['properties'][field_name]['anyOf'][0]['items'] = generate_json_objs(
                    all_schema_obj[field_prop['items']['originalRef']],
                    all_schema_obj,
                    False)
            else:
                full_schema['properties'][field_name]['anyOf'][0]['items'] = get_data_type_and_format(
                    field_prop['items']['type'],
                    field_prop.get('format', None),
                    obj_type_nullable=False)
        else:
            if 'originalRef' in field_prop:
                full_schema['properties'][field_name] = generate_json_objs(all_schema_obj[field_prop['originalRef']],
                                                                           all_schema_obj)
            else:
                full_schema['properties'][field_name] = get_data_type_and_format(field_prop['type'],
                                                                                 field_prop.get('format', None))
    full_schema['properties'] = {k: v for k, v in sorted(full_schema['properties'].items())}
    return full_schema


def generate_json_schema(stream_name, file_path):
    stream_data = client.request(method='GET',
                                 path=f'swagger/{file_path}',
                                 version='remove_accept_header')

    stream_fields_w_refs = stream_data['definitions']
    json_stream_name_form = convert_snake_to_pascal(streams.get_stream_json_obj_form(stream_name))

    if json_stream_name_form not in stream_fields_w_refs:
        raise StreamJsonObjectNotFound(f'Json object for "{stream_name}" stream not found. '
                                       f'Available objects: {", ".join(stream_fields_w_refs.keys())}')

    stream_schema = generate_json_objs(stream_fields_w_refs[json_stream_name_form], stream_fields_w_refs, False)

    stream_schema['properties'].update(CUSTOM_FIELDS_FIELD)
    return stream_schema


def main():
    if os.path.exists(OUTPUT_DIR_PATH):
        shutil.rmtree(OUTPUT_DIR_PATH)
    os.mkdir(OUTPUT_DIR_PATH)

    for stream_name, file_path in get_stream_resource_file_paths():
        file_dir_path = os.path.join(OUTPUT_DIR_PATH, f'{streams.convert_swaggered_to_tap_stream(stream_name)}.json')
        with open(file_dir_path, 'w') as f:
            json.dump(generate_json_schema(stream_name, file_path), f, indent=2)

    created_files = [file_name for file_name in os.listdir(OUTPUT_DIR_PATH) if
                     os.path.isfile(os.path.join(OUTPUT_DIR_PATH, file_name))]

    if len(created_files) != streams.get_tap_streams_count():
        created_files_name_only = [created_file.replace('.json', '') for created_file in created_files]
        for tap_stream in streams.get_tap_streams_name():
            if tap_stream not in created_files_name_only:
                raise FileNotFoundError(f'Json Schema for "{tap_stream}" stream not found')


if __name__ == '__main__':
    main()
